from osbot_fast_api.api.routes.Fast_API__Routes           import Fast_API__Routes
from mgraph_ai_service_cache.service.cache.Cache__Service import Cache__Service

TAG__ROUTES_SERVER                  = 'server'
ROUTES_PATHS__SERVER                = []       # f'/{TAG__ROUTES_SERVER}/namespaces' ]

class Routes__Server(Fast_API__Routes):
    tag         : str          = TAG__ROUTES_SERVER
    cache_service : Cache__Service

    # # todo: fix the logic, since this will only return the active namespaces in memory (and not the ones from the cloud storage
    # def namespaces(self) -> Dict[str, Any]:                                        # List all active namespaces
    #     namespaces = self.cache_service.list_namespaces()
    #     return {"namespaces": [str(ns) for ns in namespaces], "count": len(namespaces)}

    def setup_routes(self):
        #self.add_route_get(self.namespaces)
        pass




from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id import Safe_Str__Id

from memory_fs.path_handlers.Path__Handler                                        import Path__Handler
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path

class Path__Handler__Semantic_File(Path__Handler):                                  # Hash-based sharding path handler for references

    def generate_path(self, file_id : Safe_Str__Id              = None,
                            file_key: Safe_Str__File__Path = None
                       ) -> Safe_Str__File__Path:                                   # Generate sharded path based on hash
        if file_id is None:
            raise ValueError('In Path__Handler__Semantic_File, file_id cannot be None')

        if file_key is None:
            raise ValueError('In Path__Handler__Semantic_File, file_key cannot be None')
        #hash_str = str(file_id)

        return self.combine_paths(file_key)


import gzip
import json
from typing                                                                              import Dict, Optional, Any, List, Literal
from osbot_utils.decorators.methods.cache_on_self                                        import cache_on_self
from osbot_utils.type_safe.Type_Safe                                                     import Type_Safe
from osbot_utils.type_safe.primitives.core.Safe_UInt                                     import Safe_UInt
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path        import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id          import Safe_Str__Id
from osbot_utils.utils.Env                                                               import get_env
from osbot_utils.utils.Files                                                             import file_extension, file_name_without_extension
from osbot_utils.utils.Http                                                              import url_join_safe
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid                    import Random_Guid
from osbot_utils.utils.Misc                                                              import timestamp_now
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.cache.consts__Cache_Service                         import DEFAULT_CACHE__NAMESPACE
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Store__Strategy            import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache.schemas.consts.const__Fast_API                              import ENV_VAR__CACHE__SERVICE__BUCKET_NAME, ENV_VAR__CACHE__SERVICE__DEFAULT_TTL_HOURS
from mgraph_ai_service_cache.service.cache.Cache__Handler                                import Cache__Handler
from mgraph_ai_service_cache.service.cache.Cache__Hash__Config                           import Cache__Hash__Config
from mgraph_ai_service_cache.service.cache.Cache__Hash__Generator                        import Cache__Hash__Generator
from mgraph_ai_service_cache.schemas.cache.Schema__Cache__Store__Response                import Schema__Cache__Store__Response
from mgraph_ai_service_cache.service.storage.Storage_FS__S3                              import Storage_FS__S3

DEFAULT__CACHE__SERVICE__BUCKET_NAME        = "mgraph-ai-cache"
DEFAULT__CACHE__SERVICE__DEFAULT_TTL_HOURS  = 24

class Cache__Service(Type_Safe):                                                   # Main cache service orchestrator
    cache_handlers    : Dict[Safe_Str__Id, Cache__Handler]                              # Multiple cache handlers by namespace
    default_bucket    : Safe_Str__Id                  = None
    default_ttl_hours : Safe_UInt                     = None
    hash_config       : Cache__Hash__Config                                        # Hash generation config
    hash_generator    : Cache__Hash__Generator                                     # Hash generator instance

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if not self.default_bucket:
            self.default_bucket = get_env(ENV_VAR__CACHE__SERVICE__BUCKET_NAME         , DEFAULT__CACHE__SERVICE__BUCKET_NAME      )
        if not self.default_ttl_hours:
            self.default_ttl_hours = get_env(ENV_VAR__CACHE__SERVICE__DEFAULT_TTL_HOURS, DEFAULT__CACHE__SERVICE__DEFAULT_TTL_HOURS)



    def delete_by_id(self, cache_id: Random_Guid, namespace: Safe_Str__Id = None) -> Dict[str, Any]:
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        with handler.fs__refs_id.file__json(Safe_Str__Id(str(cache_id))) as ref_fs:
            if not ref_fs.exists():
                return {"status": "not_found", "message": f"Cache ID {cache_id} not found"}

            id_ref_data  = ref_fs.content()
            all_paths    = id_ref_data.get("all_paths", {})
            cache_hash   = id_ref_data.get("hash")
            strategy     = id_ref_data.get("strategy")

        # Track deletion results
        deleted_paths = []
        failed_paths  = []

        # Delete data files first (use the appropriate fs based on strategy)
        fs_data = handler.get_fs_for_strategy(strategy)
        for path in all_paths.get("data", []):
            try:
                if fs_data.storage_fs.file__delete(path):
                    deleted_paths.append(path)
                else:
                    failed_paths.append(path)
            except Exception as e:
                failed_paths.append(f"{path}: {str(e)}")

        # Update hash reference (remove this cache_id from the list)
        if cache_hash:
            with handler.fs__refs_hash.file__json(Safe_Str__Id(cache_hash)) as ref_fs:
                if ref_fs.exists():
                    refs = ref_fs.content()
                    # Remove this cache_id from the list
                    refs["cache_ids"] = [entry for entry in refs["cache_ids"]
                                        if entry["id"] != str(cache_id)]
                    refs["total_versions"] -= 1

                    if refs["total_versions"] > 0:
                        # Update the latest_id if needed
                        if refs["latest_id"] == str(cache_id) and refs["cache_ids"]:
                            refs["latest_id"] = refs["cache_ids"][-1]["id"]
                        ref_fs.update(file_data=refs)
                    else:
                        # No more versions, delete the hash reference files
                        # FIX: Changed from all_paths.get("refs", {}).get("hash", [])
                        for path in all_paths.get("by_hash", []):
                            try:
                                if handler.fs__refs_hash.storage_fs.file__delete(path):
                                    deleted_paths.append(path)
                                else:
                                    failed_paths.append(path)
                            except Exception as e:
                                failed_paths.append(f"{path}: {str(e)}")

        # Finally, delete the ID reference files
        # FIX: Changed from all_paths.get("refs", {}).get("id", [])
        for path in all_paths.get("by_id", []):
            try:
                if handler.fs__refs_id.storage_fs.file__delete(path):
                    deleted_paths.append(path)
                else:
                    failed_paths.append(path)
            except Exception as e:
                failed_paths.append(f"{path}: {str(e)}")

        return {
            "status"        : "success" if not failed_paths else "partial",
            "cache_id"      : str(cache_id),
            "deleted_count" : len(deleted_paths),
            "failed_count"  : len(failed_paths),
            "deleted_paths" : deleted_paths,
            "failed_paths"  : failed_paths
        }

    def get_all_namespaces_stats(self) -> Dict[str, Any]:       # Get file counts for all active namespaces
        all_stats = {}

        for namespace in self.cache_handlers.keys():
            counts_data = self.get_namespace__file_counts(namespace)
            all_stats[str(namespace)] = {
                'total_files': counts_data['total_files'],
                'file_counts': counts_data['file_counts']
            }

        return {
            'namespaces': all_stats,
            'total_namespaces': len(self.cache_handlers),
            'grand_total_files': sum(ns['total_files'] for ns in all_stats.values())
        }

    def get_namespace__file_hashes(self, namespace: Safe_Str__Id ) -> Dict[str, Any]:
        file_hashes = []
        parent_folder = url_join_safe(str(namespace), "refs/by-hash")

        for file_path in self.storage_fs().folder__files__all(parent_folder=parent_folder):
            if file_extension(file_path) == '.json':
                file_id = file_name_without_extension(file_path)
                file_hashes.append(file_id)
        return file_hashes

    def get_namespace__file_ids(self, namespace: Safe_Str__Id ) -> Dict[str, Any]:
        file_ids = []
        parent_folder = url_join_safe(str(namespace), "refs/by-id")

        for file_path in self.storage_fs().folder__files__all(parent_folder=parent_folder):
            if file_extension(file_path) == '.json':
                file_id = file_name_without_extension(file_path)
                file_ids.append(file_id)
        return file_ids

    def get_namespace__file_counts(self, namespace: Safe_Str__Id = None) -> Dict[str, Any]:       # Get file counts for all strategies in a namespace
        namespace = namespace or Safe_Str__Id("default")
        handler = self.get_or_create_handler(namespace)

        file_counts = {}
        total_files = 0

        # todo: review the performance implications of this on large namespaces
        for strategy in ["direct", "temporal", "temporal_latest", "temporal_versioned"]:    # Count files in each data strategy
            try:
                fs = handler.get_fs_for_strategy(strategy)
                if fs and fs.storage_fs:
                    count = len(fs.storage_fs.files__paths())
                    file_counts[f"{strategy}_files"] = count
                    total_files += count
                else:
                    file_counts[f"{strategy}_files"] = 0
            except Exception:
                file_counts[f"{strategy}_files"] = 0

        # Count reference store files
        try:
            refs_hash_count = len(handler.fs__refs_hash.storage_fs.files__paths())
            file_counts['refs_hash_files'] = refs_hash_count
            total_files += refs_hash_count
        except Exception:
            file_counts['refs_hash_files'] = 0

        try:
            refs_id_count = len(handler.fs__refs_id.storage_fs.files__paths())
            file_counts['refs_id_files'] = refs_id_count
            total_files += refs_id_count
        except Exception:
            file_counts['refs_id_files'] = 0

        file_counts['total_files'] = total_files

        return {
            'namespace': str(namespace),
            'handler': handler,
            'file_counts': file_counts,
            'total_files': total_files
        }

    def get_or_create_handler(self, namespace: Safe_Str__Id = DEFAULT_CACHE__NAMESPACE  # Get existing or create new cache handler
                              ) -> Cache__Handler:
        if namespace not in self.cache_handlers:
            handler = Cache__Handler(s3__bucket      = self.default_bucket,
                                     s3__prefix      = str(namespace),
                                     cache_ttl_hours = self.default_ttl_hours).setup()
            self.cache_handlers[namespace] = handler
        return self.cache_handlers[namespace]

    @cache_on_self
    def storage_fs(self) -> Storage_FS__S3:
        return Storage_FS__S3(s3_bucket=self.default_bucket).setup()

    def store_with_strategy(self, storage_data     : Any                                   ,
                                  cache_hash       : Safe_Str__Cache_Hash                  ,
                                  strategy         : Enum__Cache__Store__Strategy          ,
                                  cache_id         : Random_Guid                  = None                       ,
                                  cache_key        : Safe_Str__File__Path         = None   ,  # Allow extra key/path to be provided (used by some path_handlers)
                                  file_id          : Safe_Str__Id                      = None   ,
                                  namespace        : Safe_Str__Id                      = None   ,
                                  content_encoding : Safe_Str__Id                      = None
                            ) -> Schema__Cache__Store__Response:
        if not cache_hash:
            raise ValueError("in Cache__Service.store_with_strategy, the cache_hash must be provided")                      # todo: see if it makes sense for use to calculate the hash here

        cache_id  = cache_id or Random_Guid()
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)
        fs_data   = handler.get_fs_for_strategy(strategy)
        all_paths = { "data": [], "by_hash": [], "by_id" : []     }

        file_id   =  Safe_Str__Id  (file_id or cache_id )               # if not provided use the cache_id as file_id (needs casting to Safe_Str__Id)
        file_key  = Safe_Str__File__Path(cache_key)                     # use cache_key as file_key
        if isinstance(storage_data, bytes):                             # Determine file type based on storage data
            file_fs = fs_data.file__binary(file_id=file_id, file_key=file_key)
            file_type = "binary"
        else:
            file_fs = fs_data.file__json(file_id=file_id, file_key=file_key)
            file_type = "json"

        # Store actual data
        with file_fs:
            all_paths['data']  = file_fs.create(storage_data)
            content_file_paths = file_fs.file_fs__paths().paths__content()          # get the file paths for the content files

            # Add metadata
            metadata = { "cache_hash"       : str(cache_hash)    ,                  # todo: convert to Type_Safe class
                         "cache_key"        : str(cache_key)     ,
                         "cache_id"         : str(cache_id)      ,
                         "content_encoding" : content_encoding   ,
                         "file_id"          : str(file_id)       ,
                         "stored_at"        : timestamp_now()    ,
                         "strategy"         : strategy           ,
                         "namespace"        : str(namespace)     ,
                         "file_type"        : file_type          }
            file_fs.metadata__update(metadata)
            file_size = file_fs.metadata().content__size

        # Update hash->ID reference
        with handler.fs__refs_hash.file__json(Safe_Str__Id(cache_hash)) as ref_fs:
            if ref_fs.exists():
                refs = ref_fs.content()
                refs["cache_ids"     ].append({"id": str(cache_id), "timestamp": timestamp_now()})
                refs["latest_id"     ] = str(cache_id)
                refs["total_versions"] += 1
                paths__hash_to_id =  ref_fs.update(file_data=refs)
            else:
                refs = {"hash"           : str(cache_hash)                                     ,
                        "cache_ids"      : [{"id": str(cache_id), "timestamp": timestamp_now()}],
                        "latest_id"      : str(cache_id)                                        ,
                        "total_versions" : 1                                                    }
                paths__hash_to_id = ref_fs.create(file_data=refs)
            all_paths["by_hash"] = paths__hash_to_id

        # Update ID->hash reference WITH content path and file type
        with handler.fs__refs_id.file__json(Safe_Str__Id(str(cache_id))) as ref_fs:
            all_paths["by_id"] = ref_fs.paths()
            paths__id_to_hash  =  ref_fs.create({ "all_paths"        : all_paths          ,
                                                   "cache_id"         : str(cache_id)     ,
                                                   "hash"             : str(cache_hash)   ,
                                                   "namespace"        : str(namespace)    ,
                                                   "strategy"         : strategy          ,
                                                   "content_paths"    : content_file_paths,
                                                   "file_type"        : file_type         ,
                                                   "timestamp"        : timestamp_now()   })

        return Schema__Cache__Store__Response(cache_id   = cache_id     ,
                                              hash       = cache_hash   ,
                                              namespace  = namespace,
                                              paths      = all_paths    ,
                                              size       = file_size    )


    def retrieve_by_hash(self, cache_hash : Safe_Str__Cache_Hash,
                               namespace  : Safe_Str__Id = None
                          ) -> Optional[Dict[str, Any]]:                        # Retrieve latest by hash"""
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        with handler.fs__refs_hash.file__json(Safe_Str__Id(cache_hash)) as ref_fs:   # Get hash->ID mapping
            if not ref_fs.exists():
                return None
            refs = ref_fs.content()
            latest_id = refs.get("latest_id")

        if not latest_id:
            return None

        return self.retrieve_by_id(Random_Guid(latest_id), namespace)           # Delegate to retrieve_by_id which handles the path lookup

    def retrieve_by_id(self, cache_id : Random_Guid,
                             namespace : Safe_Str__Id = None
                        ) -> Optional[Dict[str, Any]]:                  #  Retrieve by cache ID using direct path from reference
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        # Get ID reference with content path
        with handler.fs__refs_id.file__json(Safe_Str__Id(cache_id)) as ref_fs:
            if not ref_fs.exists():
                return None
            ref_data = ref_fs.content()

        content_paths = ref_data.get("content_paths", [])
        file_type     = ref_data.get("file_type", "json")

        if not content_paths:
            return None

        # Get the storage backend
        storage = handler.fs__refs_id.storage_fs

        # Read the content file directly (first path is the main content)
        content_path = content_paths[0] if content_paths else None

        if content_path and storage.file__exists(content_path):
            if file_type == "binary":
                data = storage.file__bytes(content_path)
            else:
                data = storage.file__json(content_path)

            # Read metadata
            metadata_path = content_path + '.metadata'              # todo: review this usage since we should have a much better way to do this using Memory_FS
            metadata_data = {}

            if storage.file__exists(metadata_path):
                metadata_raw = storage.file__json(metadata_path)
                # Convert metadata keys to Safe_Str__Id for consistency
                metadata_data = metadata_raw.get('data')            # todo: use native Memory_FS methods here (and we should be using a Type_Safe class here)

            # Get content encoding from metadata
            content_encoding = metadata_data.get('content_encoding')

            # Handle decompression if needed
            if content_encoding == 'gzip' and isinstance(data, bytes):
                data = gzip.decompress(data)
                # After decompression, determine if it's JSON or remains binary
                try:                                            # todo: we should know this from the metadata/config
                    data = json.loads(data.decode('utf-8'))
                    data_type = "json"
                except (json.JSONDecodeError, UnicodeDecodeError):
                    data_type = "binary"
            else:
                data_type = self._determine_data_type(data)

            return { "data"            : data             ,                 # convert to Type_Safe class
                     "metadata"        : metadata_data    ,
                     "data_type"       : data_type        ,
                     "content_encoding": content_encoding }

        return None

    #todo: this method should return Schema__Cache__Entry__Details (which should be the class used to save the file)
    def retrieve_by_id__config(self, cache_id  : Random_Guid,
                                     namespace : Safe_Str__Id    = None
                                ) -> Optional[Dict[str, Any]]:                      #   Retrieve by cache ID using direct path from reference
        namespace = namespace or Safe_Str__Id("default")
        handler   = self.get_or_create_handler(namespace)

        with handler.fs__refs_id.file__json(Safe_Str__Id(cache_id)) as ref_fs:           # get the main by-id file, which contains pointers to the other files
            if not ref_fs.exists():
                return None
            return ref_fs.content()

    # todo: check who is using this
    def _is_binary_data(self, metadata) -> bool:                                    # Check if stored data is binary based on metadata
        if not metadata:
            return False

        content_encoding = metadata.data.get(Safe_Str__Id('content_encoding'))      # Check for content encoding or binary indicators
        if content_encoding:
            return True

        # Could add more checks here based on content_type if we store it
        return False

    # todo: we should be using the Enum for this
    def _determine_data_type(self, data) -> str:                                    # Determine the type of data (string, json, binary)
        if isinstance(data, bytes):
            return "binary"
        elif isinstance(data, dict) or isinstance(data, list):
            return "json"
        else:
            return "string"

    def hash_from_string(self, data: str) -> Safe_Str__Cache_Hash:                       # Calculate hash from string
        return self.hash_generator.from_string(data)

    def hash_from_bytes(self, data: bytes) -> Safe_Str__Cache_Hash:                      # Calculate hash from bytes
        return self.hash_generator.from_bytes(data)

    def hash_from_json(self, data          : dict        ,                               # Calculate hash from JSON
                             exclude_fields : List[str] = None
                        ) -> Safe_Str__Cache_Hash:
        return self.hash_generator.from_json(data, exclude_fields)


from typing                                                                       import Dict, Any, Optional
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text      import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from mgraph_ai_service_cache.schemas.errors.Schema__Cache__Error__Base            import Schema__Cache__Error__Base

class Schema__Cache__Error__Invalid_Input(Schema__Cache__Error__Base):                # 400 Bad Request errors
    field_name    : Safe_Str__Id                                                      # Which field had invalid input
    field_value   : Safe_Str__Text     = None                                         # The invalid value (if safe to include)
    expected_type : Safe_Str__Id                                                     # What type/format was expected
    constraints   : Dict[str, Any]     = None                                         # Any constraints that were violated



from osbot_utils.type_safe.Type_Safe                                                import Type_Safe
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash    import Safe_Str__Cache_Hash
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid              import Random_Guid


class Schema__Cache__Retrieve__Request(Type_Safe):  # Request schema for retrieving cache data
    hash            : Safe_Str__Cache_Hash    = None
    cache_id        : Random_Guid       = None
    include_data    : bool              = True
    include_metadata: bool              = True
    include_config  : bool              = True


from typing                                                                       import Dict, Any, Optional, Union
from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from mgraph_ai_service_cache.schemas.cache.Schema__Cache__Metadata                import Schema__Cache__Metadata
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Data_Type           import Enum__Cache__Data_Type

# Success response for retrieving data
class Schema__Cache__Retrieve__Success(Type_Safe):        # Successful retrieval with data
    data      : Union[dict, str, bytes]                   # Actual cached content
    metadata  : Schema__Cache__Metadata                   # Cache metadata
    data_type : Enum__Cache__Data_Type                    # Type of data returned



from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.cache.Schema__Cache__Metadata                import Schema__Cache__Metadata

# todo: see refactoring opportunity for creating some base classes that have common fields like cache_id, cache_hash and metadata
# Response for checking if cache entry exists
class Schema__Cache__Exists__Response(Type_Safe):                                     # Response for exists check
    exists       : bool                                                               # Whether entry exists
    cache_hash   : Safe_Str__Cache_Hash                                               # Hash that was checked
    namespace    : Safe_Str__Id                                                       # Namespace checked
    cache_id     : Random_Guid             = None                                     # ID if it exists
    metadata     : Schema__Cache__Metadata = None                                     # Metadata if it exists



from fastapi                                                                      import Response
from typing                                                                       import List, Dict, Any
from osbot_fast_api.api.decorators.route_path                                     import route_path
from osbot_fast_api.api.routes.Fast_API__Routes                                   import Fast_API__Routes
from osbot_utils.decorators.methods.cache_on_self                                 import cache_on_self
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now  import Timestamp_Now
from osbot_utils.utils.Json                                                       import bytes_to_json, json_to_str
from mgraph_ai_service_cache.schemas.cache.consts__Cache_Service                  import DEFAULT__HTTP_CODE__FILE_NOT_FOUND
from mgraph_ai_service_cache.service.cache.Cache__Service                         import Cache__Service

TAG__ROUTES_STORAGE   = 'admin/storage'

# todo:review the use of namespace in path, since it creates an inconsistency without adding a lot of value
ROUTES_PATHS__STORAGE = [ f'/{TAG__ROUTES_STORAGE}/bucket-name'                          ,  # Current bucket name
                          f'/{TAG__ROUTES_STORAGE}/file/exists/{{path:path}}'            ,  # File Exists
                          f'/{TAG__ROUTES_STORAGE}/file/bytes/{{path:path}}'             ,  # File contents (as bytes)
                          f'/{TAG__ROUTES_STORAGE}/file/json/{{path:path}}'              ,  # File contents (as json)
                          f'/{TAG__ROUTES_STORAGE}/files/parent-path'                    ,  # Files Parent Path
                          f'/{TAG__ROUTES_STORAGE}/files/all/{{path:path}}'              ,  # Files All Path
                          f'/{TAG__ROUTES_STORAGE}/folders'                              ,  # Folders
                          f'/{TAG__ROUTES_STORAGE}/{{path:path}}'                        ]  # File Delete

# todo: move to different Fast_API server/endpoint and add admin authorization
class Routes__Admin__Storage(Fast_API__Routes):
    tag           : str          = TAG__ROUTES_STORAGE
    cache_service : Cache__Service

    @cache_on_self
    def storage_fs(self):
        return self.cache_service.storage_fs()

    def bucket_name(self):
        return {'bucket-name': self.storage_fs().s3_bucket }

    @route_path("/file/exists/{path:path}")
    def file__exists(self, path):
        exists = self.storage_fs().file__exists(path)
        return dict(exists = exists,
                    path  =  path )

    @route_path("/file/bytes/{path:path}")
    def file__bytes(self, path):
        file_bytes = self.storage_fs().file__bytes(path)
        return Response(file_bytes, media_type='application/octet-stream')

    @route_path("/file/json/{path:path}")
    def file__json(self, path):
        file_bytes = self.storage_fs().file__bytes(path)
        if file_bytes:
            file_json   = bytes_to_json(file_bytes)
            return file_json
        error_data    = dict( error_type = "FILE_NOT_FOUND" ,
                              resource   = "file"           ,
                              message    = "The requested file does not exist in storage",
                              path       = path             )

        return Response(json_to_str(error_data)                           ,
                        status_code = DEFAULT__HTTP_CODE__FILE_NOT_FOUND  ,          # using 404 to indicate file not found
                        media_type  = 'application/json')

    def files__parent_path(self, path            : Safe_Str__File__Path = ''   ,
                                 return_full_path: bool                 = False
                            ) -> List:
        folders = self.storage_fs().folder__files(folder_path=path, return_full_path=return_full_path)
        return folders

    # todo: as it is this could have major performance implications (for large namespaces
    #       this is also a good example of the kind of method/capability that we could run on a schedule and return the cached version

    @route_path("/files/all/{path:path}")
    def files__all__path(self, path     : Safe_Str__File__Path = ''
                          ) -> dict:
        files      = self.storage_fs().folder__files__all(parent_folder=path)
        file_count = len(files)
        return dict(timestamp  = Timestamp_Now(),
                    file_count = file_count     ,
                    files      = sorted(files)  )

    def folders(self, path            : Safe_Str__File__Path = ''   ,
                      return_full_path: bool                 = False
                 ) -> List:
        folders = self.storage_fs().folder__folders(parent_folder=path, return_full_path=return_full_path)
        return folders

    @route_path('/{path:path}')
    def delete__file(self, path : Safe_Str__File__Path) -> Dict[str, Any]:
        return self.cache_service.storage_fs().file__delete(path)


    def setup_routes(self):
        self.add_route_get   (self.bucket_name             )
        self.add_route_get   (self.file__exists            )
        self.add_route_get   (self.file__bytes             )
        self.add_route_get   (self.file__json              )
        self.add_route_get   (self.files__parent_path      )
        self.add_route_get   (self.files__all__path        )
        self.add_route_get   (self.folders                 )

        self.add_route_delete(self.delete__file    )



from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id     import Safe_Str__Id
from memory_fs.path_handlers.Path__Handler                                          import Path__Handler
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path   import Safe_Str__File__Path

class Path__Handler__Hash_Sharded(Path__Handler):                                  # Hash-based sharding path handler for references
    shard_depth : int = 2                                                          # Two levels: aa/bb/
    shard_size  : int = 2                                                          # 2 chars per level

    def generate_path(self, file_id : Safe_Str__Id         = None                   # allow the file_id to be used by overwritten methods
                          , file_key: Safe_Str__File__Path = None                   # allow the file_key to be used by overwritten methods
                       ) -> Safe_Str__File__Path:                                   # Generate sharded path based on hash
        if file_id is None:
            raise ValueError('In Path__Handler__Hash_Sharded, file_id cannot be None')
        hash_str = str(file_id)

        shards = []                                                                 # Generate sharded path: ab/cd/abcdef123456.json
        for i in range(self.shard_depth):
            start = i * self.shard_size
            end   = start + self.shard_size
            if start < len(hash_str):
                shards.append(hash_str[start:end])

        sharded_path = '/'.join(shards) if shards else ''
        return self.combine_paths(Safe_Str__File__Path(sharded_path))


from memory_fs.Memory_FS                                                                import Memory_FS
from memory_fs.helpers.Memory_FS__Temporal                                              import Memory_FS__Temporal
from memory_fs.helpers.Memory_FS__Latest_Temporal                                       import Memory_FS__Latest_Temporal
from osbot_utils.type_safe.Type_Safe                                                    import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path       import Safe_Str__File__Path
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Store__Strategy           import Enum__Cache__Store__Strategy
from mgraph_ai_service_cache.service.storage.Storage_FS__S3                             import Storage_FS__S3
from mgraph_ai_service_cache.service.storage.path_handlers.Path__Handler__Hash_Sharded  import Path__Handler__Hash_Sharded
from mgraph_ai_service_cache.service.storage.path_handlers.Path__Handler__Semantic_File import Path__Handler__Semantic_File

# Constants for all prefix paths
CACHE__HANDLER__PREFIX_PATH__FS__REFS_ID                            = 'refs/by-id'
CACHE__HANDLER__PREFIX_PATH__FS__REFS_HASH                          = 'refs/by-hash'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_DIRECT                        = 'data/direct'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_SEMANTIC_FILE                 = 'data/semantic-file'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL                      = 'data/temporal'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_LATEST               = 'data/temporal-latest'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED            = 'data/temporal-versioned'
CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED_VERSIONS   = 'data/temporal-versioned/versions'

# Default shard depth for hash sharding
CACHE__HANDLER__SHARD_DEPTH_DEFAULT = 2

class Cache__Handler(Type_Safe):                                                   # Single handler supporting all storage strategies
    s3__bucket              : str
    s3__prefix              : str                         = ""
    cache_ttl_hours         : int                         = 24

    # Storage backend
    s3__storage             : Storage_FS__S3              = None

    # All Memory_FS instances for different strategies
    fs__data_direct             : Memory_FS                   = None                   # Direct to hash location
    fs__data_temporal           : Memory_FS__Temporal         = None                   # Temporal organization
    fs__data_temporal_latest    : Memory_FS__Latest_Temporal  = None                   # Temporal + latest
    fs__data_temporal_versioned : Memory_FS                   = None                   # Temporal + latest + versions
    fs__data_semantic_file      : Memory_FS                   = None                   # For Semantic Files

    # Reference stores (always needed)
    fs__refs_hash           : Memory_FS                   = None                    # Hash->ID mappings
    fs__refs_id             : Memory_FS                   = None                    # ID->Hash mappings

    def setup(self) -> 'Cache__Handler':                                            # Setup all storage strategies
        self.s3__storage = Storage_FS__S3(s3_bucket=self.s3__bucket,
                                          s3_prefix=self.s3__prefix).setup()

        # Reference stores (always needed)
        with Memory_FS() as _:                                                                          # Hash->ID mappings
            kwargs__fs__refs_hash = dict(prefix_path = CACHE__HANDLER__PREFIX_PATH__FS__REFS_HASH,
                                         shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = self.s3__storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__refs_hash))
            self.fs__refs_hash = _

        with Memory_FS() as _:                                                                          # ID->Hash mappings
            kwargs__fs__refs_id = dict(prefix_path = CACHE__HANDLER__PREFIX_PATH__FS__REFS_ID,
                                       shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = self.s3__storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__refs_id))
            self.fs__refs_id = _

        with Memory_FS() as _:                                                                          # Direct strategy
            kwargs__fs__data_direct = dict(prefix_path = CACHE__HANDLER__PREFIX_PATH__FS__DATA_DIRECT,
                                           shard_depth = CACHE__HANDLER__SHARD_DEPTH_DEFAULT)
            _.storage_fs = self.s3__storage
            _.add_handler(Path__Handler__Hash_Sharded(**kwargs__fs__data_direct))
            self.fs__data_direct = _

        prefix_path = Safe_Str__File__Path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL)              # Temporal only
        with Memory_FS__Temporal(storage_fs=self.s3__storage, prefix_path=prefix_path) as _:
            self.fs__data_temporal = _

        prefix_path = Safe_Str__File__Path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_LATEST)       # Temporal + Latest
        with Memory_FS__Latest_Temporal(storage_fs=self.s3__storage, prefix_path=prefix_path) as _:
            self.fs__data_temporal_latest = _

        # Temporal + Latest + Versioned
        with Memory_FS() as _:
            _.storage_fs = self.s3__storage
            _.add_handler__temporal (prefix_path = Safe_Str__File__Path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED))
            _.add_handler__latest   (prefix_path = Safe_Str__File__Path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED))
            _.add_handler__versioned(prefix_path = Safe_Str__File__Path(CACHE__HANDLER__PREFIX_PATH__FS__DATA_TEMPORAL_VERSIONED_VERSIONS))
            self.fs__data_temporal_versioned = _

        # Semantic File
        with Memory_FS() as _:                                                                          # Direct strategy
            kwargs__fs__data_direct = dict(prefix_path = CACHE__HANDLER__PREFIX_PATH__FS__DATA_SEMANTIC_FILE)
            _.storage_fs = self.s3__storage
            _.add_handler(Path__Handler__Semantic_File(**kwargs__fs__data_direct))
            self.fs__data_semantic_file = _
        return self

    def get_fs_for_strategy(self, strategy: Enum__Cache__Store__Strategy
                             ) -> Memory_FS:                                                                         # Return appropriate Memory_FS for strategy
        if strategy == "direct":
            return self.fs__data_direct
        elif strategy == "temporal":
            return self.fs__data_temporal
        elif strategy == "temporal_latest":
            return self.fs__data_temporal_latest
        elif strategy == "temporal_versioned":
            return self.fs__data_temporal_versioned
        elif strategy == "semantic_file":
            return self.fs__data_semantic_file
        else:
            raise ValueError(f"Unknown strategy: {strategy}")


from osbot_fast_api.api.routes.Routes__Set_Cookie                      import Routes__Set_Cookie
from osbot_fast_api_serverless.fast_api.Serverless__Fast_API           import Serverless__Fast_API
from mgraph_ai_service_cache.config                                    import FAST_API__TITLE
from mgraph_ai_service_cache.fast_api.routes.Routes__Delete            import Routes__Delete
from mgraph_ai_service_cache.fast_api.routes.Routes__Exists            import Routes__Exists
from mgraph_ai_service_cache.fast_api.routes.Routes__Info              import Routes__Info
from mgraph_ai_service_cache.fast_api.routes.Routes__Namespace         import Routes__Namespace
from mgraph_ai_service_cache.fast_api.routes.Routes__Retrieve          import Routes__Retrieve
from mgraph_ai_service_cache.fast_api.routes.Routes__Server            import Routes__Server
from mgraph_ai_service_cache.fast_api.routes.Routes__Admin__Storage    import Routes__Admin__Storage
from mgraph_ai_service_cache.fast_api.routes.Routes__Store             import Routes__Store
from mgraph_ai_service_cache.utils.Version                             import version__mgraph_ai_service_cache



class Service__Fast_API(Serverless__Fast_API):
    title   = FAST_API__TITLE
    version = version__mgraph_ai_service_cache

    def setup_routes(self):
        self.add_routes(Routes__Admin__Storage)
        self.add_routes(Routes__Store         )
        self.add_routes(Routes__Retrieve      )
        self.add_routes(Routes__Exists        )
        self.add_routes(Routes__Delete        )
        self.add_routes(Routes__Namespace     )
        self.add_routes(Routes__Server        )
        self.add_routes(Routes__Info          )
        self.add_routes(Routes__Set_Cookie    )
        #self.add_routes(Routes__Cache     )         # to remove one all methods have been refactored out

    # def setup_middlewares(self):
    #     super().setup_middlewares()
    #
    #     return self


# # todo: see if we need to refactor this into Fast_API since the current routes are working ok for this project (for string, bytes and dict)
# intercept the bytes submitted here
#self.app().add_middleware(BodyReaderMiddleware)

# from fastapi import Request
# from starlette.middleware.base import BaseHTTPMiddleware
# class BodyReaderMiddleware(BaseHTTPMiddleware):
#     async def dispatch(self, request: Request, call_next):
#         if request.method in ["POST", "PUT", "PATCH"]:                              # Only read body for certain methods
#             body = await request.body()
#             request.state.body = body                                               # Store it in request state for later sync access
#         else:
#             request.state.body = None
#
#         response = await call_next(request)
#         return response





from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.errors.Schema__Cache__Error__Base            import Schema__Cache__Error__Base


class Schema__Cache__Error__Conflict(Schema__Cache__Error__Base):                    # 409 Conflict errors
    conflict_type : Safe_Str__Id                                                     # Type of conflict (duplicate, locked, etc.)
    existing_id   : Random_Guid           = None                                     # Existing entry that conflicts
    cache_hash    : Safe_Str__Cache_Hash = None                                      # Hash involved in conflict



from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.core.Safe_UInt                              import Safe_UInt
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text      import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.cache.Schema__Cache__Metadata                import Schema__Cache__Metadata
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Data_Type           import Enum__Cache__Data_Type

# Response when binary data can't be included in JSON
class Schema__Cache__Binary__Reference(Type_Safe):      # Reference to binary data
    message      : Safe_Str__Text                       # Explanation message
    data_type    : Enum__Cache__Data_Type               # Always BINARY
    size         : Safe_UInt                            # Size of binary data
    cache_hash   : Safe_Str__Cache_Hash                 # Hash of the data
    cache_id     : Random_Guid                          # ID of the entry
    namespace    : Safe_Str__Id                         # Namespace
    binary_url   : Safe_Str__File__Path                 # URL to retrieve binary data
    metadata     : Schema__Cache__Metadata              # Full metadata





from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text      import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now           import Timestamp_Now


# todo: should we add here the status_code value?
class Schema__Cache__Error__Base(Type_Safe):           # Base error response schema that all error responses inherit from
    error_type  : Safe_Str__Id                         # Type of error (e.g., NOT_FOUND, INVALID_INPUT)
    message     : Safe_Str__Text                       # Human-readable error message
    timestamp   : Timestamp_Now                        # When the error occurred
    request_id  : Random_Guid                          # Unique ID for this request (for tracing)


from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.core.Safe_UInt                              import Safe_UInt


class Schema__Cache__Namespaces__List(Type_Safe):                                     # List of namespaces
    namespaces : list                                                                 # List of namespace names
    count      : Safe_UInt                                                            # Total count



from typing                                                                       import List, Dict
from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id

from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash


class Schema__Cache__Store__Response(Type_Safe):
    cache_id  : Random_Guid
    hash      : Safe_Str__Cache_Hash
    namespace : Safe_Str__Id
    paths     : Dict[str,List[Safe_Str__File__Path]]                     # Structured paths
    size      : int                                                      # Size in bytes


# import base64
# import json
# from typing                                                                        import Dict, Any, Literal
# from fastapi                                                                       import Request, Response, Body
# from osbot_fast_api.api.routes.Fast_API__Routes                                    import Fast_API__Routes
# from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                   import Safe_Str__Cache_Hash
# from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid              import Random_Guid
# from mgraph_ai_service_cache.service.cache.Cache__Service                          import Cache__Service
# from mgraph_ai_service_cache.schemas.cache.Schema__Cache__Store__Response          import Schema__Cache__Store__Response
#
# TAG__ROUTES_CACHE                  = 'cache'
# ROUTES_PATHS__CACHE                = [ f'/{TAG__ROUTES_CACHE}' + '/delete/by-id/{cache_id}/{namespace}'             ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/exists/{cache_hash}/{namespace}'                 ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/store/binary/{strategy}/{namespace}'             ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/store/json/{strategy}/{namespace}'               ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/store/string/{strategy}/{namespace}'             ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/by-hash/{cache_hash}/{namespace}'       ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/by-id/{cache_id}/{namespace}'           ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/stats/namespaces/{namespace}'                    ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/stats/namespaces'                                ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/namespaces'                                      ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/string/by-id/{cache_id}/{namespace}'    ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/json/by-id/{cache_id}/{namespace}'      ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/binary/by-id/{cache_id}/{namespace}'    ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/string/by-hash/{cache_hash}/{namespace}',
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/json/by-hash/{cache_hash}/{namespace}'  ,
#                                        f'/{TAG__ROUTES_CACHE}' + '/retrieve/binary/by-hash/{cache_hash}/{namespace}']
#
# class Routes__Cache(Fast_API__Routes):                                             # FastAPI routes for cache operations
#     tag           : str            = TAG__ROUTES_CACHE
#     cache_service : Cache__Service
#
#     def delete__by_id__cache_id__namespace(self, cache_id: Random_Guid,
#                                                  namespace: Safe_Str__Id = None) -> Dict[str, Any]:
#         return self.cache_service.delete_by_id(cache_id, namespace)
#
#     def store__string__strategy__namespace(self, #request   : Request,
#                                                  data: str = Body(...)      ,
#                                                  strategy  : Literal["direct", "temporal", "temporal_latest", "temporal_versioned"] = "temporal",
#                                                  namespace : Safe_Str__Id = None
#                                             ) -> Schema__Cache__Store__Response:
#
#         #data = request.state.body.decode()
#         cache_hash = self.cache_service.hash_from_string(data)
#         cache_id   = Random_Guid()
#
#         return self.cache_service.store_with_strategy(cache_key_data = data      ,
#                                                       storage_data   = data      ,
#                                                       cache_hash     = cache_hash,
#                                                       cache_id       = cache_id  ,
#                                                       strategy       = strategy  ,
#                                                       namespace      = namespace )
#
#     def store__json__strategy__namespace(self, data            : dict                                                                               ,
#                                                strategy        : Literal["direct", "temporal", "temporal_latest", "temporal_versioned"] = "temporal",
#                                                namespace       : Safe_Str__Id = None
#                    ) -> Schema__Cache__Store__Response:
#         #exclude_fields = []
#         # Calculate hash from filtered JSON
#         #cache_key_json = {k: v for k, v in data.items() if k not in (exclude_fields or [])}
#         #data           = str_to_json(request.state.body.decode())
#         #data           = str_to_json(data)
#         cache_hash     = self.cache_service.hash_from_json(data)
#         cache_id       = Random_Guid()
#
#         # Store full JSON
#         return self.cache_service.store_with_strategy(cache_key_data = data         ,       # todo: review this since the fact that we are storing the entire data could cause performance problems
#                                                       storage_data   = data         ,
#                                                       cache_hash     = cache_hash   ,
#                                                       cache_id       = cache_id     ,
#                                                       strategy       = strategy     ,
#                                                       namespace      = namespace    )
#
#
#     def store__binary__strategy__namespace(self, request  : Request                                                  , # todo: see if we really need this (since this is mainly used to handle gzip )
#                                                  body     : bytes = Body(..., media_type="application/octet-stream") ,
#                                                  strategy : Literal["direct", "temporal", "temporal_latest", "temporal_versioned"] = "temporal",
#                                                  namespace: Safe_Str__Id = None
#                                          ) -> Schema__Cache__Store__Response:               # Store raw binary data with hash calculation
#         #body = request.state.body
#
#         # Check if compressed
#         content_encoding = request.headers.get('content-encoding')
#         if content_encoding == 'gzip':
#             import gzip
#             decompressed   = gzip.decompress(body)
#             cache_hash     = self.cache_service.hash_from_bytes(decompressed)
#             storage_data   = body                       # Store compressed
#             cache_key_data = cache_hash                 # Use the hash of decompressed data as cache key
#         else:
#             cache_hash     = self.cache_service.hash_from_bytes(body)
#             storage_data   = body
#             cache_key_data = cache_hash                 # Store hash instead of raw binary
#
#         cache_id = Random_Guid()
#
#         return self.cache_service.store_with_strategy(cache_key_data   = str(cache_key_data),  # Convert hash to string for storage
#                                                       storage_data     = storage_data,
#                                                       cache_hash       = cache_hash,
#                                                       cache_id         = cache_id,
#                                                       strategy         = strategy,
#                                                       namespace        = namespace,
#                                                       content_encoding = content_encoding)
#
#     # todo: this should not be done like this (we should be able to get from the headers the info if the json data is compressed
#     # async def store_json_compressed(self, request        : Request                ,  # Store compressed JSON
#     #                                      exclude_fields  : List[str] = None       ,
#     #                                      strategy        : Literal["direct", "temporal", "temporal_latest", "temporal_versioned"] = "temporal",
#     #                                      namespace       : Safe_Str__Id = None
#     #                                 ) -> Schema__Cache__Store__Response:
#     #     import gzip
#     #     import json
#     #
#     #     body         = await request.body()
#     #     decompressed = gzip.decompress(body)
#     #     data         = json.loads(decompressed)
#     #
#     #     # Calculate hash from filtered, uncompressed JSON
#     #     cache_key_json = {k: v for k, v in data.items() if k not in (exclude_fields or [])}
#     #     cache_hash     = self.cache_service.hash_from_json(cache_key_json)
#     #     cache_id       = Random_Guid()
#     #
#     #     # Store compressed version
#     #     return self.cache_service.store_with_strategy(cache_key_data   = cache_key_json,
#     #                                                   storage_data     = body          ,  # Store original compressed
#     #                                                   cache_hash       = cache_hash    ,
#     #                                                   cache_id         = cache_id      ,
#     #                                                   strategy         = strategy      ,
#     #                                                   namespace        = namespace     ,
#     #                                                   content_encoding = 'gzip'         )
#
#     # def hash_calculate(self, data          : str = None                           ,  # Calculate hash from provided data
#     #                          json_data      : dict = None                          ,
#     #                          exclude_fields : List[str] = None
#     #                     ) -> Dict[str, str]:
#     #     if data:
#     #         hash_value = self.cache_service.hash_from_string(data)
#     #     elif json_data:
#     #         hash_value = self.cache_service.hash_from_json(json_data, exclude_fields)
#     #     else:
#     #         return {"error": "No data provided"}
#     #
#     #     return {"hash": str(hash_value)}
#
#     def exists__cache_hash__namespace(self, cache_hash      : Safe_Str__Cache_Hash                                   ,  # Check if hash exists
#                                             namespace  : Safe_Str__Id = None
#                                        ) -> Dict[str, bool]:
#         namespace = namespace or Safe_Str__Id("default")
#         handler   = self.cache_service.get_or_create_handler(namespace)
#
#         with handler.fs__refs_hash.file__json(Safe_Str__Id(str(cache_hash))) as ref_fs:
#             exists = ref_fs.exists()
#
#         return {"exists": exists, "hash": str(cache_hash)}
#
#     def namespaces(self) -> Dict[str, Any]:                                        # List all active namespaces
#         namespaces = self.cache_service.list_namespaces()
#         return {"namespaces": [str(ns) for ns in namespaces], "count": len(namespaces)}
#
#     def stats__namespaces__namespace(self, namespace: Safe_Str__Id = None) -> Dict[str, Any]:       # Get cache statistics
#         namespace = namespace or Safe_Str__Id("default")
#
#         try:
#             # Get file counts using shared method
#             counts_data = self.cache_service.get_namespace_file_counts(namespace)
#             handler = counts_data['handler']
#
#             # Build stats response
#             stats = { "namespace": str(namespace)         ,
#                       "s3_bucket": handler.s3__bucket     ,
#                       "s3_prefix": handler.s3__prefix     ,
#                       "ttl_hours": handler.cache_ttl_hours,
#                       **counts_data['file_counts'        ]}  # Spread all the file counts
#
#             return stats
#         except Exception as e:
#             return {"error": str(e), "namespace": str(namespace)}
#
#     def stats__namespaces(self):
#         return self.cache_service.list_namespaces()
#
#     def retrieve__by_hash__cache_hash__namespace(self, cache_hash : Safe_Str__Cache_Hash,
#                                                        namespace  : Safe_Str__Id = None
#                                                   ) -> Dict[str, Any]:                          # Retrieve latest by hash with type information"""
#         result = self.cache_service.retrieve_by_hash(cache_hash, namespace)
#         if result is None:
#             return {"status": "not_found", "message": "Cache entry not found"}
#
#         if result.get("data_type") == "binary":                                                 # If data is binary, don't return it in JSON - direct to binary endpoint
#             return { "status"      : "binary_data"                                                          ,
#                      "message"     : "Binary data cannot be returned in JSON response"                      ,
#                      "data_type"   : "binary"                                                               ,
#                      "size"        : len(result.get("data", b""))                                           ,
#                      "metadata"    : result.get("metadata", {})                                             ,
#                      "binary_url"  : f"/cache/retrieve/binary/by-hash/{cache_hash}/{namespace or 'default'}"}
#
#         return result
#
#     def retrieve__by_id__cache_id__namespace(self, cache_id  : Random_Guid,
#                                                    namespace : Safe_Str__Id = None
#                                               ) -> Dict[str, Any]:              # Retrieve by cache ID with type information
#         result = self.cache_service.retrieve_by_id(cache_id, namespace)
#         if result is None:
#             return {"status": "not_found", "message": "Cache entry not found"}
#
#
#         if result.get("data_type") == "binary":                                 # If data is binary, don't return it in JSON - direct to binary endpoint
#             return { "status"      : "binary_data"                                                      ,
#                      "message"     : "Binary data cannot be returned in JSON response"                  ,
#                      "data_type"   : "binary"                                                           ,
#                      "size"        : len(result.get("data", b""))                                       ,
#                      "metadata"    : result.get("metadata", {})                                         ,
#                      "binary_url"  : f"/cache/retrieve/binary/by-id/{cache_id}/{namespace or 'default'}" }
#
#         return result
#
#     # New type-specific retrieval methods
#     def retrieve__string__by_id__cache_id__namespace(self, cache_id: Random_Guid,
#                                                     namespace: Safe_Str__Id = None
#                                                     ) -> str:
#         """Retrieve as string by cache ID"""
#         result = self.cache_service.retrieve_by_id(cache_id, namespace)
#         if result is None:
#             return Response(content="Not found", status_code=404)
#
#         data = result.get("data")
#         data_type = result.get("data_type")
#
#         if data_type == "string":
#             return Response(content=data, media_type="text/plain")
#         elif data_type == "json":
#             import json
#             return Response(content=json.dumps(data), media_type="text/plain")
#         elif data_type == "binary":
#             # Convert binary to string (might not be ideal for all binary data)
#             try:
#                 return Response(content=data.decode('utf-8'), media_type="text/plain")
#             except:
#                 # If can't decode, return base64 encoded
#                 return Response(content=base64.b64encode(data).decode('utf-8'),
#                               media_type="text/plain")
#
#         return Response(content=str(data), media_type="text/plain")
#
#     def retrieve__json__by_id__cache_id__namespace(self, cache_id: Random_Guid,
#                                                   namespace: Safe_Str__Id = None
#                                                   ) -> Dict[str, Any]:
#         """Retrieve as JSON by cache ID"""
#         result = self.cache_service.retrieve_by_id(cache_id, namespace)
#         if result is None:
#             return {"status": "not_found", "message": "Cache entry not found"}
#
#         data = result.get("data")
#         data_type = result.get("data_type")
#
#         if data_type == "json":
#             return data
#         elif data_type == "string":
#             # Try to parse string as JSON
#             try:
#                 import json
#                 return json.loads(data)
#             except:
#                 return {"error": "Data is not valid JSON", "data": data}
#         elif data_type == "binary":
#             # Return base64 encoded binary in JSON wrapper
#             return {
#                 "data_type": "binary",
#                 "encoding": "base64",
#                 "data": base64.b64encode(data).decode('utf-8')
#             }
#
#         return {"data": data, "data_type": data_type}
#
#     def retrieve__binary__by_id__cache_id__namespace(self, cache_id: Random_Guid,
#                                                     namespace: Safe_Str__Id = None):     # Retrieve as binary by cache ID
#         result = self.cache_service.retrieve_by_id(cache_id, namespace)
#         if result is None:
#             return Response(content="Not found", status_code=404)
#
#         data      = result.get("data")
#         data_type = result.get("data_type")
#
#         if data_type == "binary":
#             return Response(content=data,
#                             media_type="application/octet-stream")          # Return raw binary data (already decompressed by service)
#         elif data_type == "string":
#
#             return Response(content=data.encode('utf-8'),
#                             media_type="application/octet-stream")          # Convert string to bytes
#         elif data_type == "json":
#
#             return Response(content=json.dumps(data).encode('utf-8'),
#                             media_type="application/octet-stream")          # Convert JSON to bytes
#
#         return Response(content=str(data).encode('utf-8'),
#                       media_type="application/octet-stream")                # Fallback
#
#     def retrieve__string__by_hash__cache_hash__namespace(self, cache_hash: Safe_Str__Cache_Hash,
#                                                         namespace: Safe_Str__Id = None):
#         """Retrieve as string by hash"""
#         result = self.cache_service.retrieve_by_hash(cache_hash, namespace)
#         if result is None:
#             return Response(content="Not found", status_code=404)
#
#         data = result.get("data")
#         data_type = result.get("data_type")
#
#         if data_type == "string":
#             return Response(content=data, media_type="text/plain")
#         elif data_type == "json":
#             import json
#             return Response(content=json.dumps(data), media_type="text/plain")
#         elif data_type == "binary":
#             try:
#                 return Response(content=data.decode('utf-8'), media_type="text/plain")
#             except:
#                 return Response(content=base64.b64encode(data).decode('utf-8'),
#                               media_type="text/plain")
#
#         return Response(content=str(data), media_type="text/plain")
#
#     def retrieve__json__by_hash__cache_hash__namespace(self, cache_hash: Safe_Str__Cache_Hash,
#                                                       namespace: Safe_Str__Id = None
#                                                       ) -> Dict[str, Any]:
#         """Retrieve as JSON by hash"""
#         result = self.cache_service.retrieve_by_hash(cache_hash, namespace)
#         if result is None:
#             return {"status": "not_found", "message": "Cache entry not found"}
#
#         data = result.get("data")
#         data_type = result.get("data_type")
#
#         if data_type == "json":
#             return data
#         elif data_type == "string":
#             try:
#                 import json
#                 return json.loads(data)
#             except:
#                 return {"error": "Data is not valid JSON", "data": data}
#         elif data_type == "binary":
#             return {
#                 "data_type": "binary",
#                 "encoding": "base64",
#                 "data": base64.b64encode(data).decode('utf-8')
#             }
#
#         return {"data": data, "data_type": data_type}
#
#     def retrieve__binary__by_hash__cache_hash__namespace(self, cache_hash: Safe_Str__Cache_Hash,
#                                                         namespace: Safe_Str__Id = None):
#         """Retrieve as binary by hash"""
#         result = self.cache_service.retrieve_by_hash(cache_hash, namespace)
#         if result is None:
#             return Response(content="Not found", status_code=404)
#
#         data = result.get("data")
#         data_type = result.get("data_type")
#         content_encoding = result.get("content_encoding")
#
#         if data_type == "binary":
#             headers = {}
#             if content_encoding:
#                 headers["Content-Encoding"] = content_encoding
#             return Response(content=data, media_type="application/octet-stream",
#                           headers=headers)
#         elif data_type == "string":
#             return Response(content=data.encode('utf-8'),
#                           media_type="application/octet-stream")
#         elif data_type == "json":
#             import json
#             return Response(content=json.dumps(data).encode('utf-8'),
#                           media_type="application/octet-stream")
#
#         return Response(content=str(data).encode('utf-8'),
#                       media_type="application/octet-stream")
#     def setup_routes(self):                                                        # Configure FastAPI routes
#         # String endpoints
#         self.add_route_post(self.store__string__strategy__namespace)
#
#         # JSON endpoints
#         self.add_route_post(self.store__json__strategy__namespace)
#
#         # Binary endpoints
#         self.add_route_post(self.store__binary__strategy__namespace)
#
#         # Generic retrieval endpoints (return with metadata and type info)
#         self.add_route_get(self.retrieve__by_hash__cache_hash__namespace)
#         self.add_route_get(self.retrieve__by_id__cache_id__namespace)
#
#         # Type-specific retrieval by ID
#         self.add_route_get(self.retrieve__string__by_id__cache_id__namespace)
#         self.add_route_get(self.retrieve__json__by_id__cache_id__namespace)
#         self.add_route_get(self.retrieve__binary__by_id__cache_id__namespace)
#
#         # Type-specific retrieval by hash
#         self.add_route_get(self.retrieve__string__by_hash__cache_hash__namespace)
#         self.add_route_get(self.retrieve__json__by_hash__cache_hash__namespace)
#         self.add_route_get(self.retrieve__binary__by_hash__cache_hash__namespace)
#
#         # Utility endpoints
#         self.add_route_get(self.exists__cache_hash__namespace)
#         self.add_route_get(self.namespaces)
#         self.add_route_get(self.stats__namespaces__namespace)
#         self.add_route_get(self.stats__namespaces)
#         self.add_route_delete(self.delete__by_id__cache_id__namespace)
#
#     # # todo: remove this method when the next version of OSBot-Fast-API has been installed (which uses the code below)
#     # def add_route_delete(self, function):
#     #     return self.add_route_with_body(function, methods=['DELETE'])


from osbot_utils.type_safe.primitives.core.Safe_UInt                              import Safe_UInt
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_int.Timestamp_Now           import Timestamp_Now
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from mgraph_ai_service_cache.schemas.errors.Schema__Cache__Error__Base            import Schema__Cache__Error__Base


class Schema__Cache__Error__Gone(Schema__Cache__Error__Base):                        # 410 Gone errors (expired entries)
    cache_id    : Random_Guid                = None                                  # ID that expired
    expired_at  : Timestamp_Now              = None                                  # When it expired
    ttl_hours   : Safe_UInt                  = None                                  # What the TTL was
    namespace   : Safe_Str__Id               = None                                  # Namespace of expired entry



from enum import Enum

class Enum__Cache__Store__Strategy(str, Enum):
    DIRECT              = "direct"
    TEMPORAL            = "temporal"
    TEMPORAL_LATEST     = "temporal_latest"
    TEMPORAL_VERSIONED  = "temporal_versioned"
    SEMANTIC_FILE       = "semantic_file"



import hashlib
import json
from typing                                                      import List
from osbot_utils.type_safe.Type_Safe                             import Type_Safe
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                      import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.service.cache.Cache__Hash__Config   import Cache__Hash__Config, Enum__Hash__Algorithm

class Cache__Hash__Generator(Type_Safe):                                           # Generate consistent hashes from various input types
    config: Cache__Hash__Config

    def calculate(self, data: bytes) -> Safe_Str__Cache_Hash:                      # Calculate hash from raw bytes
        if self.config.algorithm == Enum__Hash__Algorithm.MD5:
            hash_full = hashlib.md5(data).hexdigest()
        elif self.config.algorithm == Enum__Hash__Algorithm.SHA256:
            hash_full = hashlib.sha256(data).hexdigest()
        elif self.config.algorithm == Enum__Hash__Algorithm.SHA384:
            hash_full = hashlib.sha384(data).hexdigest()

        return Safe_Str__Cache_Hash(hash_full[:self.config.length])

    def from_string(self, data: str) -> Safe_Str__Cache_Hash:                       # Hash from string
        return self.calculate(data.encode('utf-8'))

    def from_bytes(self, data: bytes) -> Safe_Str__Cache_Hash:                      # Hash from bytes
        return self.calculate(data)

    def from_json(self, data      : dict     ,                                      # Hash JSON with optional field exclusion
                        exclude_fields: List[str] = None
                   ) -> Safe_Str__Cache_Hash:
        if exclude_fields:
            data = {k: v for k, v in data.items() if k not in exclude_fields}
        json_str = json.dumps(data, sort_keys=True)
        return self.from_string(json_str)

    def from_type_safe(self, obj           : Type_Safe     ,                      # Hash Type_Safe object
                            exclude_fields : List[str] = None
                       ) -> Safe_Str__Cache_Hash:
        return self.from_json(obj.json(), exclude_fields)


from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Store__Strategy import Enum__Cache__Store__Strategy

DEFAULT_CACHE__STORE__STRATEGY     = Enum__Cache__Store__Strategy.TEMPORAL
DEFAULT_CACHE__NAMESPACE          = 'default'

DEFAULT__HTTP_CODE__FILE_NOT_FOUND = 404


import re

from osbot_utils.type_safe.primitives.core.Safe_Str                         import Safe_Str
from osbot_utils.type_safe.primitives.core.enums.Enum__Safe_Str__Regex_Mode import Enum__Safe_Str__Regex_Mode

TYPE_SAFE_STR__GITHUB__SHA_SHORT__REGEX  = re.compile(r'^[a-fA-F0-9]{7}$')
TYPE_SAFE_STR__GITHUB__SHA_SHORT__LENGTH = 7

# todo: move to OSBot_Utils
class Safe_Str__SHA1__Short(Safe_Str):
    """
    Safe string class for short 7-character Git commit SHAs.

    Examples:
    - "7fd1a60"
    - "abc1234"
    """
    regex             = TYPE_SAFE_STR__GITHUB__SHA_SHORT__REGEX
    regex_mode        = Enum__Safe_Str__Regex_Mode.MATCH
    max_length        = TYPE_SAFE_STR__GITHUB__SHA_SHORT__LENGTH
    exact_length      = True
    allow_empty       = True
    trim_whitespace   = True
    strict_validation = True



from typing                                                                       import Dict, List
from osbot_utils.type_safe.Type_Safe                                              import Type_Safe
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text      import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.files.safe_str.Safe_Str__File__Path import Safe_Str__File__Path
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.identifiers.safe_str.Safe_Str__Id   import Safe_Str__Id
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Data_Type           import Enum__Cache__Data_Type
from mgraph_ai_service_cache.schemas.cache.enums.Enum__Cache__Store__Strategy     import Enum__Cache__Store__Strategy

# Response for cache entry details
class Schema__Cache__Entry__Details(Type_Safe):                                       # Detailed cache entry info
    cache_id       : Random_Guid                                                      # Cache ID
    cache_hash     : Safe_Str__Cache_Hash                                             # Content hash
    namespace      : Safe_Str__Id                                                     # Namespace
    strategy       : Enum__Cache__Store__Strategy                                     # Storage strategy
    all_paths      : Dict[Safe_Str__Id, List[Safe_Str__File__Path]]                   # All file paths
    content_paths  : List[Safe_Str__File__Path]                                       # Content file paths
    file_type      : Enum__Cache__Data_Type                                           # Type of data
    timestamp      : Safe_Str__Text                                                   # When stored


from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text      import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.identifiers.Random_Guid             import Random_Guid
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                       import Safe_Str__Cache_Hash
from mgraph_ai_service_cache.schemas.errors.Schema__Cache__Error__Base            import Schema__Cache__Error__Base


class Schema__Cache__Error__Unsupported_Media_Type(Schema__Cache__Error__Base):       # 415 Unsupported Media Type errors
    requested_type : Safe_Str__Text                                                   # Type that was requested
    actual_type    : Safe_Str__Text                                                   # Actual type of the data
    cache_hash     : Safe_Str__Cache_Hash = None                                      # Hash of the entry (if applicable)
    cache_id       : Random_Guid          = None                                      # ID of the entry (if applicable)



from typing                                                                              import Dict, List
from osbot_utils.type_safe.Type_Safe                                                     import Type_Safe
from osbot_utils.type_safe.primitives.domains.http.safe_str.Safe_Str__Http__Content_Type import Safe_Str__Http__Content_Type
from osbot_utils.type_safe.primitives.domains.common.safe_str.Safe_Str__Text             import Safe_Str__Text
from osbot_utils.type_safe.primitives.domains.cryptography.safe_str.Safe_Str__Cache_Hash                                              import Safe_Str__Cache_Hash


class Schema__Cache__Store__Request(Type_Safe):          # Request schema for storing cache data"""
    data        : str                                    # Base64 encoded for binary data   # todo: see what type of schema we should use here
    content_type: Safe_Str__Http__Content_Type           # MIME type                        # todo: see if we need this here
    metadata    : Dict[Safe_Str__Id, Safe_Str__Text]          # Domain-specific metadata
    tags        : List[Safe_Str__Id]                          # Tags for categorization
    hash        : Safe_Str__Cache_Hash                         # provide custom hash              #todo: see if it is a good idea to ask the called to provide the hash
